=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed April 2008 by Fredo6

# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted, provided that the above
# copyright notice appear in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name		:   OffsetOnSurface.rb
# Type		:   Sketchup Tool
# Description	:   offset a contour on a surface (inside and outside)
# Menu Items	:   Tools --> "Offset on Surface"
# Context Menu	:   none
# Usage		:   See Tutorial in PDF format
# Original Date	:   10 April 2008
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

require 'sketchup.rb'
require 'LibTraductor.rb'			# for language translation

module OffsetOnSurface

STR_Title = ["Offset on surface",
	         "|FR| Offset de contour sur une surface"]
			 
STR_Tooltip = ["Offset outer edges of any surface",	
	           "|FR| Agrandit ou retrecit un contour sur une surface"]	 
STR_MsgInvalidSelection = ["NO Contour for the selection",
                           "|FR| Pas de contour d\fini pour la s\lection"]
STR_MsgComplexSelection = ["Selection too complex for offset operation",
                           "|FR| S\lection trop complexe pour l'op\ration"]

MSG_MnuDone = "Done |FR| Termin\"
MSG_MnuGroup = [ "Generate as Group", "|FR| G\n\rer comme Groupe"]
MSG_MnuAlone = [ "Isolate surface when extending", "|FR| Traiter comme surface isol\e"]
MSG_MnuContours = [ "Contour Selection", "|FR| S\lection Contours"]
MSG_MnuNoFaces = [ "Generate faces", "|FR| G\n\rer faces"]
MSG_MnuNoCurves = [ "Generate as curves", "|FR| G\n\rer comme courbe"]
MSG_MnuCurrent = [ "(current = ", "|FR| (courant = "]


MSG_Distance = "Distance |FR| Distance"
MSG_Face_Origin = ["Select Surface to offset",
                    "|FR| S\lectionner la surface"]
MSG_Face_End = ["Drag selected edge of the face or type distance",
                 "|FR| Pousser / Tirer le bord de la face ou entrer une distance"]
MSG_Input_Execution = ["Double Click or Enter to launch",
                    "|FR| Double cliquez ou touche Enter pour ex\cuter"]
MSG_Group = ["GROUP",
             "|FR| GROUPE"]
MSG_Alone = ["STANDALONE",
             "|FR| SURFACE ISOLEE"]
MSG_Contours = ["CONTOURS",
                "|FR| CONTOURS"]
MSG_NoFaces = ["GEN. NO FACES",
                "|FR| PAS DE FACES"]
MSG_NoCurves = ["NO CURVES",
                "|FR| PAS DE COURBE"]
MSG_AllContours = ["ALL",
                   "|FR| TOUS"]
MSG_InnerContours = ["INNER",
                   "|FR| INTERIEURS"]
MSG_OuterContours = ["OUTER",
                   "|FR| EXTERIEURS"]
				   
DLG_EnumYesNo = { 'Y' => "Yes |FR| Oui", 
				  'N' => "No |FR| Non" }			 
DLG_EnumContours = { 'A' => MSG_AllContours, 'O' => MSG_OuterContours, 'I' => MSG_InnerContours }			 
				  
#Constants for OffsetOnSurface Module (do not translate)	
OFS_TOOLBAR = "NUL Transformation"
OFS_DIR = "OFS_Dir"
SU_MAJOR_VERSION_6 = (Sketchup.version[0..0] > '5')	
OFS_ICON_MAIN = "OfsMain.png"
OFS_CURSOR_OK = "OfsMain.png"
OFS_CURSOR_FORBIDDEN = "OfsForbidden.png"
OFS_CURSOR_DRAG = "OfsDrag.png"
OFS_MENU_SU = "Tools"

STATE_ORIGIN = 0
STATE_END = 1
STATE_EXECUTION = 2

ERROR_NO_SELECTION = 1
ERROR_INVALID_SELECTION = 2
ERROR_COMPLEX_SELECTION = 3

OFS___Contours = 116	#F5 - Contours: All, Inner, Outer			   
OFS___Group = 117		#F6 - Group  option			   
OFS___Alone = 118		#F7 - Isolated surface option			   
OFS___NoFaces = 119		#F8 - Do not generate face			   
OFS___NoCurves = 120	#F9- Do not map original curves			   

OFS_EdgeData = Struct.new("OFS_EdgeData", :edge, :face_in, :normal_in, :face_out, :normal_out,
                                          :pt1, :pt2, :curved, :looped, :vd_start, :vd_end, :iloop) 
OFS_VertexData = Struct.new("OFS_VertexData", :origin, :vertex, :lstedges, :newpt, :colinear, :cos_angle, 
											  :trueborder, :vxs_in, :vxs_out, :vxs, :mark, :next_in_loop,
											  :edloop, :iloop, :angle_in) 
OFS_VexSubData = Struct.new("OFS_VexSubData", :vec, :distance, :vnorm, :same_plane,
                                              :vplane, :dfactor, :chemin, :camino) 
OFS_Node = Struct.new("OFS_Node", :pt, :edge, :distance, :face, :vec) 
OFS_Mark = Struct.new("OFS_Mark", :pt, :face, :vertex, :edge, :signature) 

#--------------------------------------------------------------------------------------------------------------
# Top Calling functions: create the class and laucnh the tool
#--------------------------------------------------------------------------------------------------------------			 				   

def OffsetOnSurface.launch
	@ofstool = OFSTool.new unless @ofstool
	Sketchup.active_model.select_tool @ofstool
end

#--------------------------------------------------------------------------------------------------------------
# Class OFS: Offset Surface Tranformation: 
#--------------------------------------------------------------------------------------------------------------			 
class OFS

attr_reader :noselection, :hsh_faces, :lst_edges, :hsh_edges

def initialize
    @model = Sketchup.active_model
	@entities = @model.active_entities
	@selection = @model.selection
	@dsnap = OFSG.dsnap
	
	reset	
end

#Reset all variables for the class OFS
def reset
	@hsh_faces = {}
	@hsh_edges = {}
	@lst_edges = []
	@lst_all_edges = []
	@lst_main_edges = []
	@hsh_selected_edges = {}
	@hsh_selected_faces = {}
	@hsh_vertices = {}
	@too_complex = false
	@all_loops = []
end

def set_option_group(group)
	@option_group = group
end

def set_option_alone(alone)
	@option_alone = alone
end

def set_option_nofaces(nofaces)
	@option_nofaces = nofaces
end

def set_option_nocurves(nocurves)
	@option_nocurves = nocurves
end
	
def set_option_contours(contours)
	@option_contours = contours
	compute_active_loops if @all_loops.length > 0
end
		
#Check the selection at entry of the tool if any	
def check_initial_selection(flgmsg)
	#checking if the selection contains faces
	reset
	@selection.each do |e|
		if e.class == Sketchup::Face
			@hsh_selected_faces[e.to_s] = e 
		elsif e.class == Sketchup::Edge
			@hsh_selected_edges[e.to_s] = e if e.faces.length > 0
		end
	end	
	
	#Evaluating the selection for edges and faces
	unless inspect_selected_edges
		reset
		if (flgmsg)
			UI.beep
			Traductor.messagebox STR_MsgComplexSelection
		end	
		return ERROR_COMPLEX_SELECTION	
	end
	
	return ERROR_NO_SELECTION if (@hsh_faces.length == 0) 	#no face selected
	
	#calculating the contour
	if (@hsh_edges.length == 0)
		reset
		if (flgmsg)
			UI.beep
			Traductor.messagebox STR_MsgInvalidSelection
		end	
		return ERROR_INVALID_SELECTION
	end	
	construct_contour_loops

	#Selection is OK
	return 0
end

#Inspect the edges that are selected to check if they define a proper selection
def inspect_selected_edges
	unless @hsh_selected_edges.length > 0
		@hsh_selected_faces.each { |key, f| @hsh_faces[f.to_s] = f }
		return calculate_edges
	end
	
	lstedges = []
	@hsh_selected_edges.each do |key, e|
		lfaces = e.faces
		if (lfaces.length == 1)
			@hsh_faces[lfaces[0].to_s] = lfaces[0]
			lstedges.push e
			next
		end	
		n = 0
		face = nil
		lfaces.each do |f|
			if @hsh_selected_faces[f.to_s]
				n += 1
				face = f
			end
		end
		if n == 1	
			@hsh_faces[face.to_s] = face
			lstedges.push e
		end	
	end
	
	#rechecking edges for complex topology
	lstedges.each do |e|
		face = nil
		n = 0
		e.faces.each do |f|
			if @hsh_faces[f.to_s]
				n += 1
				face = f
			end
		end		
		store_edge_data(e, face) if n == 1	
	end
	
	#Calculating extra properties at vertices - Reject operation if too complex
	return false if @too_complex
	@hsh_vertices.each { |key, vd| properties_at_vertex vd }
	true
end

#Define a selection starting from a face with all connected faces that share a Soft edge
def virtual_selection(face)
	return true if @hsh_faces && @hsh_faces[face.to_s]
	reset
	face_neighbours face
	unless calculate_edges
		reset
		return false
	end	
	construct_contour_loops
	(@hsh_faces.length > 0) && (@hsh_edges.length > 0)
end

#Determine all connected faces to the face (i.e. if bording edge is soft or hidden)
#note: the recursive version seems to bugsplat on big number of faces. So I use an iterative version
def face_neighbours(face)
	lface = [face]
	
	while true
		break if lface.length == 0
		f = lface[0]
		if @hsh_faces[f.to_s]
			lface[0..0] = []
			next
		end	
		lface[0..0] = []
		@hsh_faces[f.to_s] = f
		f.edges.each do |e|
			if e.hidden? || e.soft?
				e.faces.each do |ff| 
					lface.push ff unless ff == f || @hsh_faces[ff.to_s]
				end	
			end	
		end
	end	
end

#Calculate the contour of the surface (i.e. edges with only one face)
def calculate_edges
	@hsh_faces.each do |key, face|
		face.edges.each do |e|
			n = 0
			e.faces.each { |f| n += 1 if @hsh_faces[f.to_s] }
			next unless (n == 1)		#edge shares more than one selected face
			
			#Storing the edges
			store_edge_data(e, face)
			return false if @too_complex
		end
	end	
	@hsh_vertices.each { |key, vd| properties_at_vertex vd }
	return true
end

def store_edge_data(e, face)
	@lst_all_edges.push e
	ed = OFS_EdgeData.new
	@hsh_edges[e.to_s] = ed
	ed.edge = e
	ed.face_in = face
	ed.face_out = nil
	ed.normal_in = OFSG.normal_ex_to_edge(e, face)
	ed.normal_out = ed.normal_in.reverse
	ed.normal_out = ed.normal_in.reverse
	e.faces.each do |f|
		if f != face
			ed.face_out = f
			ed.normal_out = OFSG.normal_ex_to_edge e, f
			break
		end
	end	
	ed.pt1 = e.start.position
	ed.pt2 = e.end.position
	ed.curved = false
	ed.looped = false
	
	#Storing the vertex data
	rev = (ed.edge.reversed_in? ed.face_in) ? true : false
	rev = false
	if rev
		ed.pt1 = e.end.position
		ed.pt2 = e.start.position
		ed.vd_start = store_vertex_data(e.end, ed)
		ed.vd_end = store_vertex_data(e.start, ed)		
	else
		ed.pt1 = e.start.position
		ed.pt2 = e.end.position
		ed.vd_start = store_vertex_data(e.start, ed)
		ed.vd_end = store_vertex_data(e.end, ed)
	end	
end

def store_vertex_data(v, ed)
	vd = @hsh_vertices[v.to_s]
	unless vd
		vd = OFS_VertexData.new
		vd.vxs_in = OFS_VexSubData.new
		vd.vxs_out = OFS_VexSubData.new
		vd.vertex = v
		vd.origin = v.position
		vd.lstedges = []
		vd.newpt = nil
		vd.mark = nil
		vd.next_in_loop = nil
		@hsh_vertices[v.to_s] = vd
	end	
	vd.lstedges.push ed
	@too_complex = true if vd.lstedges.length > 2		#complex vertex
	vd
end

#Compute and store extra properties at vertex linking 2 edges
def properties_at_vertex(vd)		
	vxs_in = vd.vxs_in
	vxs_out = vd.vxs_out
	v = vd.vertex
	origin = vd.origin

	ed1 = vd.lstedges[0]
	face1 = ed1.face_in
	
	#termination vertex
	if (vd.lstedges.length == 1)
		ed1 = vd.lstedges[0]
		vxs_in.vec = ed1.normal_in
		vxs_out.vec = ed1.normal_out
		vnorm = ed1.face_in.normal
		vxs_in.vplane = [origin, vxs_in.vec * vnorm]
		vnorm = ed1.face_out.normal if ed1.face_out
		vd.trueborder = (ed1.face_out) ? false : true
		vxs_out.vplane = [origin, vxs_out.vec * vnorm]
		#vxs_out.vplane = vxs_in.vplane
		vd.colinear = false
		vd.cos_angle = 0.0
		vxs_in.dfactor = 1.0
		vxs_out.dfactor = 1.0
		vxs_in.same_plane = true
		vxs_out.same_plane = true
		edge1 = ed1.edge
		vxs_in.camino = Camino.new(vxs_in.vplane, origin, face1, edge1, vxs_in.vec.reverse, nil)
		vxs_out.camino = Camino.new(vxs_out.vplane, origin, face1, edge1, vxs_in.vec, [face1])
		return
	end	
	
	#vertex with 2 edges
	ed1 = vd.lstedges[0]
	ed2 = vd.lstedges[1]

	#Computing parameters for Inside
	vec1 = ed1.normal_in
	vec2 = ed2.normal_in	
	vxs_in.vec = Geom.linear_combination 0.5, vec1, 0.5, vec2
	face1 = ed1.face_in
	face2 = ed2.face_in
	vxs_in.same_plane = face1.normal.parallel?(face2.normal)
	vnorm = OFSG.average_normal(face1, face2)
	vxs_in.vplane = [origin, vxs_in.vec * vnorm]
	angle = vxs_in.vec.angle_between vec1
	angle1 = vxs_in.vec.angle_between vec1
	angle2 = vxs_in.vec.angle_between vec2
	cosinus = (Math::cos(angle1) + Math::cos(angle2)) * 0.5
	vxs_in.dfactor = (cosinus.abs < 0.05) ? 1.0 : 1.0 / cosinus

	#Computing parameters for outside
	vec1 = ed1.normal_out
	vec2 = ed2.normal_out	
	vxs_out.vec = OFSG.average_vector(vec1, vec2)
	vd.trueborder = (ed1.face_out || ed2.face_out) ? false : true
	f1 = (ed1.face_out) ? ed1.face_out : face1
	f2 = (ed2.face_out) ? ed2.face_out : face2	
	vxs_out.same_plane = f1.normal.parallel?(f2.normal)
	vnorm = OFSG.average_normal(f1, f2)
	vxs_out.vplane = [origin, vxs_out.vec * vnorm]
	angle1 = vxs_out.vec.angle_between vec1
	angle2 = vxs_out.vec.angle_between vec2
	cosinus = (Math::cos(angle1) + Math::cos(angle2)) * 0.5
	vxs_out.dfactor = (cosinus.abs < 0.05) ? 1.0 : 1.0 / cosinus
	
	edge1 = ed1.edge
	edge2 = ed2.edge
	v1 = edge1.start.position.vector_to edge1.end.position
	v2 = edge2.start.position.vector_to edge2.end.position
	vd.colinear = v1.parallel? v2
	vd.cos_angle = (Math::cos v1.angle_between(v2)).abs
	vd.angle_in = convex_concave_angle origin, vxs_in.vec, edge1
	
	#creating the caminos for the vertex
	vxs_in.camino = Camino.new(vxs_in.vplane, origin, face1, edge1, vxs_in.vec.reverse, nil)
	vxs_out.camino = Camino.new(vxs_out.vplane, origin, face1, edge1, vxs_out.vec.reverse, [face1, face2])
end

def convex_concave_angle(pt, vec_in, edge)
	ptend = edge.start.position
	ptend = edge.end.position if ptend == pt
	vec = pt.vector_to ptend
	vec.angle_between(vec_in.reverse)	
end

def compute_all_vertices(distance)
	@hsh_vertices.each do |key, vd| 
		compute_vertex vd, distance 
	end	
end

def compute_vertex(vd, distance)	
	if (distance > 0)
		vd.vxs = vxs = vd.vxs_out
		if (vxs.vec.valid?)
			if @option_alone || vd.trueborder
				pt = vd.origin.offset vd.vxs_in.vec, vd.vxs_in.dfactor * distance
				vd.mark = OFSG.create_mark(pt, nil)
			elsif vd.colinear		#skip vertex that are on edges, as they usually mess up the contour
				vd.mark = OFSG.create_mark(nil, vd.edloop.face_in)
			else	
				vd.mark = vxs.camino.pursue_chemin(vxs.dfactor * distance, @dsnap)
			end	
		else
			vd.mark = OFSG.create_mark(vd.vertex.position, nil)
		end
	else
		vd.vxs = vxs = vd.vxs_in
		if vd.colinear		#skip vertex that are on edges, as they usually mess up the contour
			vd.mark = OFSG.create_mark(nil, nil)
		elsif @option_alone
			pt = vd.origin.offset vxs.vec.reverse, -vxs.dfactor * distance
			vd.mark = OFSG.create_mark(pt, nil)
		else
			vd.mark = vxs.camino.pursue_chemin(-vxs.dfactor * distance, @dsnap)
		end	
	end	
	vd.newpt = vd.mark.pt.clone if vd.mark.pt
	vd.mark.signature = vd
end

def loop_from_edge(ed, fstart)
	loop = []
	vdbeg = (fstart) ? ed.vd_start : ed.vd_end
	ed.looped = true unless fstart
	return loop unless vdbeg
	ednext = ed
	vdnext = vdbeg
	loop.push vdbeg
	vdbeg.edloop = ed
	
	while true
		break if (vdnext.lstedges.length == 1)
		ednext = (ednext == vdnext.lstedges[0]) ? vdnext.lstedges[1] : vdnext.lstedges[0]
		vdnext = (vdnext == ednext.vd_start) ? ednext.vd_end : ednext.vd_start
		ednext.looped = true
		break unless vdnext
		loop.last.next_in_loop = vdnext
		loop.push vdnext
		vdnext.edloop = ednext
		break if (vdnext == vdbeg)
	end	
	loop
end

#Organize the contours of the selection into loops
def construct_contour_loops
	@all_loops = []	
	@hsh_edges.each do |key, ed|
		next if ed.looped
		lp1 = loop_from_edge(ed, true)
		if (ed.looped)
			loop = lp1
		else	
			lp2 = loop_from_edge(ed, false)
			loop = lp1.reverse + lp2
		end	
		@all_loops.push loop
	end	

	@all_loops.each_with_index do |loop, iloop|
		loop.each do |vd|
			vd.iloop = iloop
			vd.edloop.iloop = iloop
		end	
	end	

	check_loops_type
	@lst_main_edges = []
	@lst_inner_edges = []
	@lst_outer_edges = []
	@hsh_edges.each do |key, ed|
		case @lst_loop_type[ed.iloop]
		when -1	
			@lst_inner_edges.push ed.edge
		when 1	
			@lst_outer_edges.push ed.edge
		else
			@lst_main_edges.push ed.edge
		end	
	end		
	compute_active_loops
end

def loop_status(loop)
	return 0 if loop.first != loop.last		#not a true loop
	sum = 0.0
	loop.each { |vd| sum += vd.angle_in }
	valcrit = Math::PI * loop.length.to_f
	d = sum * 2.0 - valcrit
	return 0 if d.abs < 1.0e-08
	(d < 0) ? 1 : -1
end

#check whether the type of all loops ( Inner or Outer) and build list of loops accordingly
def check_loops_type
	nb = @all_loops.length - 1
	@lst_loop_type = []
	for i in 0..nb
		@lst_loop_type.push 0
	end	
	
	@main_loops = []
	@inner_loops = []
	@outer_loops = []
		
	@all_loops.each_with_index do |loop, iloop|
		case status = loop_status(loop)
		when 1
			@outer_loops.push loop
		when -1
			@inner_loops.push loop
		else
			@main_loops.push loop
		end
		@lst_loop_type[iloop] = status
	end	
end

#Compute the active list of loops
def compute_active_loops
	@loops = @main_loops
	@lst_edges = @lst_main_edges
	opt = @option_contours
	opt = 'A' if @all_loops.length == 1
	case opt
	when 'O'
		@loops += @outer_loops
		@lst_edges += @lst_outer_edges
	when 'I'
		@loops += @inner_loops
		@lst_edges += @lst_inner_edges
	else
		@loops += @outer_loops + @inner_loops
		@lst_edges += @lst_outer_edges + @lst_inner_edges
	end	
end

#Top function to execute the Offset on Surface
def execute(distance)
	#Computing the transformed position of the contour vertices
	compute_all_vertices distance

	#performing drawing of edge
	pbar = ProgressionBar.new @hsh_edges.length, "Edges"
	@model.start_operation Traductor[STR_Title]
		grp = @entities.add_group
		@loops.each do |loop|
			lst_contour = []
			calculate_contour(loop, lst_contour, @dsnap)
			generate_all_faces grp, loop unless distance < 0 || @option_nofaces
			
			if (@option_nocurves)
				grp.entities.add_edges lst_contour if lst_contour.length > 1
			else	
				g = grp.entities.add_group
				g.entities.add_curve lst_contour
				g.explode
			end	
			#lst_contour.each { |p| grp.entities.add_cpoint p }
		end
		
		unless @option_group
			ls = grp.explode
		end	
	@model.commit_operation  
	reset
end
	
def get_new_contour(distance)
	#computing the new position of vertices
	@distance = distance
	compute_all_vertices distance

	#Simplifying the contour and joining the new position of vertices
	lpt = []
	@loops.each do |loop|		
		lst_contour = []
		calculate_contour(loop, lst_contour, @dsnap)
		lpt.push lst_contour
	end	
	lpt
end

def calculate_contour(loop, lst_contour, dsnap)

	#Simplifying the contour by removing overlapping points
	simplify_contour loop
	
	#joining the remaining points
	lp = []
	loop.each { |vd| lp.push vd if vd.mark.pt }
	parcours = []
	nv = lp.length - 2
	for i in 0..nv
		vd1 = lp[i]
		vd2 = lp[i+1]
		pk = junction vd1, vd2
		pk.each { |mk| parcours.push mk }
	end	
	if (loop.first == loop.last) && (lp.first != lp.last)
		pk = junction lp.last, lp.first
		pk.each { |mk| parcours.push mk }
	end
	
	#Cleaning up the resulting contour
	parcours = remove_dups_in_parcours(parcours)
	parcours = smooth_parcours(parcours)
	parcours = snap_in_parcours(parcours, dsnap)
	parcours.each { |mk| lst_contour.push mk.pt }
end

#Simplify the contour by removing transformed vertices that are outside the contour
def simplify_contour(loop)
	remove_dups_in_contour(loop, @dsnap)
	
	#checking whether to apply the simplification
	if (@distance > 0 && @outer_loops.include?(loop)) ||
	   (@distance < 0 && @inner_loops.include?(loop))
	   return
	end 
	#return
	
	#Eliminating redundant vertices
	correction = true
	vdprev = loop.last
	while correction
		correction = false
		loop.each do |vd|
			next unless vd.mark.pt
			unless vdprev.mark.pt
				vdprev = vd
				next
			end
			if OFSG.same_side_of_plane?(vdprev.vxs_in.vplane, vd.origin, vd.mark.pt)
				vdprev = vd
				next
			end	
			correction = true
			if vdprev.angle_in <= vd.angle_in
				vd.mark.pt = nil
			else
				vdprev.mark.pt = nil
				vdprev = vd
			end
		end	
	end	
	remove_dups_in_contour(loop, 0)
end

#remove points which are very close in a contour
def remove_dups_in_contour(loop, dsnap)
	lp = []
	loop.each { |vd| lp.push vd if vd.mark.pt }

	vdprev = lp[0]
	nb = lp.length - 1
	for i in 1..nb
		vd1 = vdprev
		vd2 = lp[i]
		next unless vd2.mark.pt && vd1.mark.pt
		if vd1.mark.pt.distance(vd2.mark.pt) <= dsnap
			if vd1.cos_angle <= vd2.cos_angle
				vd2.mark.pt = nil
			else
				vd1.mark.pt = nil
				vdprev = vd2
			end
			next
		end
		vdprev = vd2
	end			
end

def snap_in_parcours(parcours, dsnap)
	return [] unless parcours.length > 0
	dsnap_v = dsnap
	dsnap_e = dsnap
	dsnap_ev = dsnap
	parcours.each_with_index do |mk, i|
		parcours[i] = OFSG.snap_to_vertex(mk.pt, mk.face, dsnap_v) unless mk.vertex || (mk.face == nil)
	end	
	parcours.each_with_index do |mk, i|
		parcours[i] = OFSG.snap_in_face(mk.pt, mk.face, nil, dsnap_e) unless mk.vertex || (mk.face == nil)
	end	

	#Eliminating dups
	parcours = remove_dups_in_parcours(parcours)

	#smoothing edge points close to vertex
	n = parcours.length - 2
	for i in 0..n
		mk1 = parcours[i]
		mk2 = parcours[i+1]
		d = mk1.pt.distance(mk2.pt)
		if mk1.edge && mk2.vertex && d <= dsnap_ev
			parcours[i] = mk2
		elsif mk2.edge && mk1.vertex && d <= dsnap_ev	
			parcours[i+1] = mk1
		end
	end	
	
	#Eliminating dups
	remove_dups_in_parcours(parcours)	
end

def remove_dups_in_parcours(parcours)
	return [] unless parcours.length > 0
	lp = []
	n = parcours.length - 2
	for i in 0..n
		mk1 = parcours[i]
		mk2 = parcours[i+1]
		next unless mk1.pt && mk2.pt
		lp.push mk1 unless mk1.pt == mk2.pt
	end	
	lp.push parcours.last unless parcours.last.pt && lp.last.pt == parcours.last.pt
	lp
end

#remove 'free' marks corresponding to original vertices, when angle is not meaningful
def smooth_parcours(parcours)
	hlsup = {}
	n = parcours.length-2
	for i in 1..n
		mk = parcours[i]
		vd = mk.signature
		next unless vd
		next if mk.vertex || mk.edge || (mk.face == nil)
		next if OFSG.which_edge(mk.face, mk.pt)
		unless vd.vxs.same_plane || vd.vxs.dfactor > 1.01 || vd.cos_angle < 0.5
			hlsup[i] = true
		end
	end	
	
	lp = []
	parcours.each_with_index { |mk, i| lp.push mk unless hlsup[i] }
	lp
end

def generate_all_faces(grp, loop)
	nb = loop.length - 1	
	k = 0
	lstfaces = []
	ptlast = nil
	while (k < nb)
		vd1 = loop[k]
		k += 1
		next if vd1.mark.face
		ptlast = vd1.mark.pt if vd1.mark.pt
		next unless ptlast
		camino1 = vd1.vxs.camino
		next if camino1 && camino1.chemin.length > 2
		for j in k..nb
			k = j
			vd2 = loop[j]
			camino2 = vd2.vxs.camino
			if vd2.mark.face || (camino2 && camino2.chemin.length > 2)
				ptlast = nil
				break
			end	
			lstfaces += generate_single_face grp, vd1, vd2, ptlast
			break
		end	
	end
	cleanup_colinear(grp, lstfaces)
end

def generate_single_face(grp, vd1, vd2, ptlast)
	newpt1 = (vd1.mark.pt) ? vd1.mark.pt : ptlast
	newpt2 = (vd2.mark.pt) ? vd2.mark.pt : ptlast
	pt1 = vd1.origin
	pt2 = vd2.origin
	face = vd2.edloop.face_in
	
	begin
		if (newpt1 == newpt2)
			newface = grp.entities.add_face [pt1, pt2, newpt1]
		else
			newface = grp.entities.add_face [pt1, pt2, newpt2, newpt1]
		end	
		transfer_face face, newface
		return [newface]
	rescue
		newface1 = grp.entities.add_face [pt1, pt2, newpt2]
		newface2 = grp.entities.add_face [newpt2, newpt1, pt1]
		transfer_face face, newface1
		transfer_face face, newface2
		return [newface1, newface2]
	end	
end		

#generic transfer for any drawing element
def transfer_drawing_element (old_entity, new_entity)
	new_entity.layer = old_entity.layer
	new_entity.material = old_entity.material
	new_entity.visible = old_entity.visible?
	new_entity.receives_shadows = old_entity.receives_shadows?
	new_entity.casts_shadows = old_entity.casts_shadows?
end

#generic transfer for any drawing element
def transfer_face (oldface, newface)
	transfer_drawing_element oldface, newface
	newface.back_material = oldface.back_material
	newface.reverse! if (newface.normal % oldface.normal < 0) 
end

#Erase all coplanar edges on the generated faces
def cleanup_colinear(grp, lstfaces)
	ldel = []
	hedge = {}
	lstfaces.each do |face|
		face.edges.each do |e|
			next if hedge[e.to_s]
			hedge[e.to_s] = true
			if (e.faces.length == 2) 
				face1 = e.faces[0]
				face2 = e.faces[1]
				if (face1.normal.parallel? face2.normal) &&
				   (face1.material == face2.material) && (face1.back_material == face2.back_material)
					ldel.push e
				else
					e.soft = true
					e.smooth = true
				end	
			end
		end	
	end
	grp.entities.erase_entities ldel
end

#tries to reconstruct curves form the original curves present in contour
def handle_curve (grp, curve, g)
	n = curve.vertices.length - 1
	lvd = []
	interrupt = false
	ibeg = -1
	curve.vertices.each_with_index do |v, i|
		vd = @hsh_vertices[v.to_s]
		interrupt = true unless vd
		ibeg = i if ibeg == -1 && vd && interrupt
		lvd.push vd
	end
	
	#rearranging the list of vertex to have only continuous pieces of curve
	if interrupt
		lvd2 = []
		for i in 0..n
			j = i + ibeg
			j = j - n - 1 if (j > n)
			lvd2.push lvd[j]
		end
		lvd = lvd2
	end
	
	#creating the curve in the separate group
	pts = []
	lvd.each do |vd|
		unless vd			
			curve = g.entities.add_curve pts if pts.length > 1
			pts = []
		else		
			pts.push vd.mark.pt if vd.mark.pt
			vd.lstedges.each { |edd| edd.curved = true }
		end
	end		
	curve = g.entities.add_curve pts if pts.length > 1
end

def rebuild_curves(grp)
	hshv = {}
	g = grp.entities.add_group
	@lst_edges.each do |edge|
		ed = @hsh_edges[edge.to_s]
		next if ed.curved
		ed.curved = true
		curve = edge.curve
		handle_curve(grp, curve, g) if curve
	end	
	g.explode
end

#Precompute the parameters and Camino for the Red point
def set_red_point(origin, face, edge)

	#determining if the red point is a vertex
	at_vertex = nil
	dmin = edge.length / 100.0
	edge.vertices.each do |v|
		if v.position.distance(origin) < dmin
			at_vertex = v
			break
		end
	end
	
	#Computing the faces in and out
	if (at_vertex)
		vd = @hsh_vertices[at_vertex.to_s]
		ed1 = vd.lstedges[0]
		ed2 = vd.lstedges[1]
		edge1 = ed1.edge
		edge2 = ed2.edge
		face1 = ed1.face_in
		face2 = ed2.face_in
		origin = at_vertex.position
	else
		lstfaces_in = [face]
		faceout = nil
		edge.faces.each { |f| faceout = f if face != f }
	end
	
	#Creating the Camino for the red point for the INSIDE direction
	vec = edge.start.position.vector_to edge.end.position 
	vplane = [origin, vec]
	normal = OFSG.normal_ex_to_edge edge, face
	@red_camino_in = Camino.new vplane, origin, face, edge, normal.reverse, nil

	#Creating the Camino for the red point for the OUTSIDE direction
	if faceout == nil || @option_alone
		normalout = normal.reverse
	else	
		normalout = OFSG.normal_ex_to_edge edge, faceout
	end	
	@red_camino_out = Camino.new vplane, origin, face, edge, normalout.reverse, [face]
end

#Calculate the transformation of the Red Point at a distance
def path_from_red_point(distance)
	camino = (distance > 0) ? @red_camino_out : @red_camino_in
	return unless camino
	d = distance.abs
	mark = camino.pursue_chemin d, 0
	pt = mark.pt
	chemin = camino.chemin	
	path = []
	chemin.each do |node| 
		break if (node.distance >= d)
		path.push node.pt 
	end	
	path.push pt
	path
end

def junction(vd1, vd2)
	mark1 = vd1.mark
	mark2 = vd2.mark
	lstmark = calculate_junction(mark1, mark2)
	unless lstmark.length < 2
		lstmark.first.signature = vd1
		lstmark.last.signature = vd2
	end
	lstmark
end

#Calculate the junction between 2 marks on the surface
def calculate_junction(mark1, mark2)
	face1 = mark1.face
	face2 = mark2.face
	
	#ignore duplicate points, if any
	return [] if mark1.pt == mark2.pt
	
	#Direct path
	if (face1 == nil && face2 == nil)
		return [mark1, mark2]
	end
	
	#One of the mark is outside surface
	if face2 == nil
		camino = join_marks(mark1, mark2)
		camino.chemin.pop if camino.end		
		return chemin_to_parcours(camino) + [mark2]
	elsif face1 == nil
		camino = join_marks(mark2, mark1)
		camino.chemin.pop if camino.end		
		return [mark1] + chemin_to_parcours(camino).reverse
	end
	
	#Both marks are on surface
	camino1 = join_marks(mark1, mark2)
	if camino1.end
		camino2 = join_marks(mark2, mark1)
		mark = camino_cross camino1, camino2
		return chemin_to_parcours(camino1) + [mark] + chemin_to_parcours(camino2).reverse
	end	
	return chemin_to_parcours(camino1)
end

#Join <mark1> toward <mark2>, assuming both are on a face
def join_marks(mark1, mark2)
	vecdir = compute_vecdir mark1, mark2
	pt1 = mark1.pt
	pt2 = mark2.pt
	face1 = mark1.face
	face2 = mark2.face
	vnorm = OFSG.average_normal(face1, face2)
	vplane = [pt1, vecdir * vnorm]

	#Computing the camino from mark1 to mark2
	camino = Camino.new vplane, pt1, face1, nil, vecdir, nil
	camino.chemin_to_target(mark2, vecdir)
	camino
end

#Extract a list of marks from the chemin of a camino
def chemin_to_parcours(camino)
	parcours = []
	camino.chemin.each { |node| parcours.push OFSG.create_mark(node.pt, node.face) }
	parcours	
end

#Compute the termination lines of 2 caminos and find the best intersection point
def camino_cross(camino1, camino2)
	chemin1 = camino1.chemin
	chemin2 = camino2.chemin
	n1 = chemin1.length - 1
	n2 = chemin2.length - 1
	line1 = [chemin1[n1-1].pt, chemin1[n1].pt]
	line2 = [chemin2[n2-1].pt, chemin2[n2].pt]
	lpt = Geom.closest_points line1, line2
	pt = Geom.linear_combination(0.5, lpt[0], 0.5, lpt[1]) 
	chemin1.pop
	chemin2.pop
	OFSG.create_mark pt, nil
end

#Compute the initial practical direction between 2 points on the surface, handling difficult cases
def compute_vecdir(mark1, mark2)
	sinus_limit = 0.05
	
	#regular vector, not too far from the plane of the face
	pt1 = mark1.pt
	vecdir = pt1.vector_to mark2.pt
	face1 = mark1.face
	sinus = Math.sin vecdir.angle_between(face1.normal)
	return vecdir if face1 == nil || sinus.abs >= sinus_limit
	
	#trying alternate face
	ent = mark1.edge
	ent = mark1.vertex unless ent
	if ent
		ent.faces.each do |f|
			next if f == face1
			sinus = Math.sin vecdir.angle_between(f.normal)
			if sinus.abs >= sinus_limit 
				mark1.face = f
				return vecdir
			end
		end
	end
	
	#Trying to compute the best direction by generating the camino to the middle point of the vertex next edge
	vd1 = mark1.signature
	return vecdir unless vd1
	vd2 = vd1.next_in_loop
	
	ed1 = vd1.lstedges[0]
	vd2.lstedges.each { |ed| ed1 = ed if vd1.lstedges.include?(ed) } if vd2
	edge1 = ed1.edge
	
	pt_start = edge1.start.position
	pt_end = edge1.end.position
	ptmid = Geom.linear_combination 0.5, pt_start, 0.5, pt_end
	
	#face1 = mark1.face
	face1 = ed1.face_in
	if vd1.vxs == vd1.vxs_in
		face = ed1.face_in
		vec = ed1.normal_in.reverse
		lsface = nil
	else
		vec = ed1.normal_out.reverse
		face = (ed1.face_out) ? ed1.face_out : ed1.face_in
		lsface = [face]
	end		
	vplane = vec * face1.normal
	camino = Camino.new(vplane, ptmid, face, edge1, vec, lsface)
	mark = camino.pursue_chemin @distance.abs, 0
	return pt1.vector_to(mark.pt)
end

end #Class OFS

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Class Camino to hold a path on a surface
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

class Camino

attr_reader :chemin, :end

def initialize(vplane, origin, face, edge, vecdir, faces_excluded)
	@vplane = vplane
	@hface = {}
	@end = false
	@chemin = []
	@initial = -1
	@vecprev = vecdir.reverse
	faces_excluded.each { |f| @hface[f.to_s] = f } if (faces_excluded)
	next_faces_from_node(face, origin, edge)
	chemin_initial(origin, edge, face, vecdir)
end

def push_node(pt, distance, edge, face, vec)
	@chemin.push OFSG.create_node(pt, distance, edge, face, vec)
end

def point_within_chemin(distance, dsnap)
	#target point is beyond the chemin
	chemin = @chemin
	pthd = chemin.last
	
	if pthd.distance < distance
		return nil unless @end
		d = distance - pthd.distance
		return OFSG.snap_in_face(pthd.pt.offset(pthd.vec, d), nil, nil, dsnap) 	
	end
	
	#Target point is within the chemin
	n = chemin.length - 1
	pthd = nil
	for i in 0..n
		pthd = chemin[i]
		return OFSG.snap_in_face(pthd.pt, pthd.face, pthd.edge, dsnap) if pthd.distance == distance
		break if (pthd.distance > distance)
	end	
	d = pthd.distance - distance
	pt = pthd.pt.offset(pthd.vec.reverse, d)
	face = pthd.face
	unless OFSG.within_face_extended?(face, pt)		#hole in a face
		pthd.edge.faces.each do |f|
			next if f == face
			if OFSG.within_face_extended?(f, pt)
				face = f
				break
			end	
		end
	end
	return OFSG.snap_in_face(pt, face, nil, dsnap)
end

#Compute the path on a face from an origin, along a vector, and up to a certain distance
#Note: origin is located within the face, possibly on an edge
def chemin_on_face(origin, face, distance, edge, vecdir)
	#computing the intersection with the edges of the face
	chemin = @chemin
	path = []
	leng = @chemin.last.distance
	dmin = (edge) ? edge.length / 100.0 : distance / 1000.0
	face.edges.each do |e|
		next if e == edge
		pt = OFSG.intersect_edge_with_plane e, @vplane
		next unless pt && pt != origin
		d = origin.distance(pt)
		vec = origin.vector_to(pt) 
		vec.normalize!
		next if @vecprev.samedirection?(vec)	#make sure we do not go backward
		angle = vec.angle_between @vecprev
		path.push OFSG.create_node(pt, d + leng, e, face, vec)
	end
	return false if path.length == 0
	path.sort! { |x, y| x.distance <=> y.distance } if path.length > 1
	
	dd = 0.0
	path.each do |p| 
		next if dd == p.distance
		dd = p.distance
		@chemin.push p 
	end
	@vecprev = @chemin.last.vec.reverse.normalize
	return true
end

def chemin_initial(origin, edge, face0, vecdir)
	#computing the intersection with the edges of the face
	chemin = @chemin
	OFSG.watchlist_clear_face
	
	#Computing the intersections on each face
	lstpath = []
	@lface.each do |face|
		vec0 = nil
		face.edges.each do |e|
			pt = OFSG.intersect_edge_with_plane e, @vplane
			next unless pt && pt != origin
			vec = origin.vector_to(pt) 
			unless vec0
				vec0 = vec
				lstpath.push [face, vec]
			else
				next if vec.samedirection?(vec0)
				lstpath.push [face, vec]
				break
			end			
		end
	end	

	#Cleaning up the intersections to keep only non colinear vectors
	lstvec = []
	lstfv = []
	
	lstpath.each do |fv|
		unless colinear_to(fv[1], lstvec) 
			lstfv.push fv
			lstvec.push fv[1]
		end
	end
	
	#Analysing the intersections
	fv_ok = nil
	@initial = lstfv.length
	case lstfv.length
	when 0
		@chemin.push OFSG.create_node(origin, 0.0, edge, face0, vecdir)
		@end = true
		@lface = []
		@vecprev = vecdir.reverse
		return
	
	when 1
		fv = lstfv[0] 
		fv_ok = fv unless vecdir.reverse.samedirection?(fv[1]) || vecdir % fv[1] < 0
		
	else
		angle0 = Math::PI
		sinus0 = 1.0
		lstfv.each do |fv|
			if fv[1] % vecdir >= 0
				sinus = Math.sin vecdir.angle_between(fv[1])
				if (sinus < sinus0)
					fv_ok = fv
					sinus0 = sinus
				end
			end	
		end
	end
	unless fv_ok
		pt = origin.offset vecdir, 10
		vec = origin.vector_to pt.project_to_plane(face0.plane)
		@end = true
		@chemin.push OFSG.create_node(origin, 0.0, edge, face0, vec)
		@lface = []
		@vecprev = vec.reverse
		return
	end	
	vec = fv_ok[1]
	@vecprev = vec.reverse
	@chemin.push OFSG.create_node(origin, 0.0, edge, face0, vec)
	@lface = [fv_ok[0]]
end

def colinear_to(vec, lstvec)
	lstvec.each { |v| return true if v.samedirection? vec }
	false
end

#compute the full path from an origin over the surface, up to a distance (optional)
#variable <chemin> should already contain at least one node
def pursue_chemin(distance, dsnap)
	#Checking if the point is not already within the chemin
	mark = point_within_chemin(distance, dsnap)
	return mark if mark
	
	#Getting the origin and other parameters
	nodelast = @chemin.last
	origin = nodelast.pt
	edge = nodelast.edge
		
	#Loop on exploring all faces and progressing on the surface	
	while true
		#No face to explore and still to go. Going outside the surface
		if @lface.length == 0
			@end = true
			return point_within_chemin(distance, dsnap)
		end	
	
		#Getting next face to treat
		face = @lface.pop
		next if @hface[face.to_s]
		@hface[face.to_s] = face
		
		#testing the path on the current face if any
		next unless chemin_on_face(origin, face, distance, edge, nil)
			
		#Checking if the node is a vertex and determining the next face
		nodelast = @chemin.last
		origin = nodelast.pt
		edge = nodelast.edge
		next_faces_from_node(face, origin, edge)
			
		#Found point within distance
		return point_within_chemin(distance, dsnap) if (chemin.last.distance > distance)
	end	
end

#Checking if the node is a vertex and determining the next face
def next_faces_from_node(face, origin, edge)
	@lface = []
	mark = OFSG.snap_in_face(origin, face, edge, 0)
	
	if mark.vertex
		mark.vertex.faces.each { |ff| @lface.push ff unless @hface[ff.to_s] }
	elsif edge
		edge.faces.each { |ff| @lface.push ff unless @hface[ff.to_s] }
	elsif mark.edge
		mark.edge.faces.each { |ff| @lface.push ff unless @hface[ff.to_s] }
	else
		@lface = [face] unless @hface[face.to_s]
	end
end


#Find the chemin to a target point <mark> - <vecdir> indicates the first direction to follow
def chemin_to_target(mark, vecdir)
	#Getting the origin and other parameters
	nodelast = @chemin.last
	origin = nodelast.pt
	edge = nodelast.edge
	distance = nodelast.distance
	mkpt = mark.pt
		
	#Loop on exploring all faces and progressing on the surface	
	while true
		#No face to explore and still to go. Going outside the surface
		if @lface.length == 0
			@end = true
			pt = nodelast.pt.offset nodelast.vec, 1
			push_node(pt, nodelast.distance + 1, nil, nil, nodelast.vec)
			return false
		end	
	
		#Getting next face to treat
		face = @lface.pop
		next if @hface[face.to_s]
		@hface[face.to_s] = face
		
		#testing the path on the current face if any
		nb_beg = @chemin.length - 1
		next unless chemin_on_face(origin, face, distance * 1.01, edge, vecdir)
		vecdir = nil
			
		#Checking if the node is a vertex and determining the next face
		nb_end = @chemin.length - 2
		nodelast = @chemin.last
		origin = nodelast.pt
		edge = nodelast.edge
		distance = nodelast.distance
			
		#Check if target has been reached
		for i in nb_beg..nb_end
			pt1 = @chemin[i].pt
			pt2 = @chemin[i+1].pt
			case OFSG.point_within_segment(mkpt, pt1, pt2)
			when 1	#close to pt1
				@chemin[i+1..(nb_end+1)] = []
				return true
			when 2	#close to pt2	
				@chemin[i+2..(nb_end+1)] = [] if i < nb_end
				return true
			when -1  #Within segment	
				@chemin[i+1..(nb_end+1)] = []
				node = @chemin[i]
				push_node(mkpt, node.distance + pt1.distance(mkpt), node.edge, node.face, node.vec)
				return true
			end
		end
		next_faces_from_node(face, origin, edge)		
	end	#end infinite loop
	false
end

end #class Camino

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# OFSG: Utility class with standalone functions
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
WCOLOR = ["blue", "green", "yellow"]

class OFSG
@@watchlist = nil	#For debug purpose
@@watchlistf = nil	#For debug purpose

#Retrieve the length for snapping, also used as precision
def OFSG.dsnap
	op = Sketchup.active_model.options["UnitsOptions"] 
	d = op["LengthSnapLength"]
	d.to_f
end

def OFSG.snap_in_face(pt, face, edge, dsnap)
	#No face
	return OFSG.mark(pt, face, nil, nil) unless face
	
	#Close to a vertex
	ent = (edge) ? edge : face
	ent.vertices.each do |v|
		return OFSG.mark(v.position, face, v, nil) if pt.distance(v.position) <= dsnap
	end
	
	#On an edge
	edges = (edge) ? [edge] : face.edges
	edges.each do |e|
		pt1 = e.start.position
		pt2 = e.end.position
		ptproj = pt.project_to_line [pt1, pt2]
		next if (pt1.vector_to(ptproj) % pt2.vector_to(ptproj) > 0)
		return OFSG.mark(ptproj, face, nil, e) if pt.distance(ptproj) <= dsnap
	end
	
	#Free point
	face = (OFSG.within_face?(face, pt)) ? face : nil
	OFSG.mark(pt, face, nil, edge)
end

def OFSG.snap_to_vertex(pt, face, dsnap)
	return OFSG.mark(pt, face, nil, nil) unless face
	
	#Close to a vertex
	face.vertices.each do |v|
		return OFSG.mark(v.position, face, v, nil) if pt.distance(v.position) <= dsnap
	end
		
	#Free point
	face = (OFSG.within_face?(face, pt)) ? face : nil
	OFSG.mark(pt, face, nil, nil)
end

#Check if a point is within a segment (excluding vertices) - <pt> is supposed to be on the segment
#return 1 if pt1, 2 if pt2, -1 if within segment and 0 is not on segment
def OFSG.point_within_segment(pt, pt1, pt2)
	return 1 if (pt.distance(pt1) <= dsnap)
	return 2 if (pt.distance(pt2) <= dsnap)
	ptproj = pt.project_to_line [pt1, pt2]
	return -1 if (pt.distance(ptproj) == 0) && (pt1.vector_to(ptproj) % pt2.vector_to(ptproj) < 0)
	return 0
end

def OFSG.watchlist_add(lpt)
	@@watchlist = [] unless @@watchlist
	lpt.each { |pt| @@watchlist.push pt }
end

def OFSG.watchlist_add_face(lface)
	@@watchlistf = [] unless @@watchlistf
	lface.each { |f| @@watchlistf.push f }
end


def OFSG.watchlist_draw(view)
	if @@watchlist
		for i in 0..(@@watchlist.length-1)
			j = i.modulo 3
			color = WCOLOR[j]
			view.draw_points @@watchlist[i], 8, 2, color
		end	
	end
	if @@watchlistf
		for i in 0..(@@watchlistf.length-1)
			j = i.modulo 3
			color = WCOLOR[j]
			pts = []
			face = @@watchlistf[i]
			face.vertices.each { |v| pts.push v.position }
			view.drawing_color = "orange"
			view.line_width = 5
			view.draw GL_LINE_STRIP, pts
		end	
	end	
end

def OFSG.watchlist_clear
	@@watchlist = [] if @@watchlist
end

def OFSG.watchlist_clear_face
	@@watchlistf = [] if @@watchlistf
end

#Find if a point if a point is on an edge of a face and return it
def OFSG.which_edge(face, pt)
	return nil unless face
	face.edges.each do |e|
		pt1 = e.start.position
		pt2 = e.end.position
		return e if (pt == pt1) || (pt == pt2)
		ptproj = pt.project_to_line [pt1, pt2]
		d = pt.distance_to_line [pt1, pt2]
		return e if (d == 0) && (ptproj.vector_to(pt1) % ptproj.vector_to(pt2) <= 0)
	end
	return nil
end

#Computing the average normal of 2 faces (regardless of their orientation) 
def OFSG.average_normal(face1, face2)
	normal1 = face1.normal.normalize
	return normal1 unless face2
	normal2 = face2.normal.normalize	
	normal2.reverse! if normal1 % normal2 < 0
	normal2 = normal1 if normal1.parallel? normal2
	Geom.linear_combination 0.5, normal1, 0.5, normal2
end

#Computing the average normal of 2 faces (regardless of their orientation) 
def OFSG.average_vector(vec1, vec2)
	return vec1 unless vec2
	return vec2 unless vec1
	vec2 = vec1 if vec1.parallel? vec2
	Geom.linear_combination 0.5, vec1.normalize, 0.5, vec2.normalize
end

#Compute the intersection of a plane and an edge
def OFSG.intersect_edge_with_plane(e, plane)
	pt_start = e.start.position
	pt_end = e.end.position
	begin
		pt = Geom.intersect_line_plane e.line, plane
	rescue
		return nil
	end	
	return nil unless pt
	
	#Check if point cross the edge
	((pt == pt_start) || (pt == pt_end) || (pt.vector_to(pt_start) % pt.vector_to(pt_end) <= 0)) ? pt : nil
end

#Create a Node structure
def OFSG.create_node(pt, distance, edge, face, vec)
	pthd = OFS_Node.new
	pthd.distance = distance
	pthd.pt = pt.clone
	pthd.edge = edge
	pthd.face = face
	pthd.vec = vec
	pthd
end

#create a Mark structure
def OFSG.create_mark(pt, face)
	mark = OFS_Mark.new
	mark.pt = (pt) ? pt.clone : nil
	mark.face = face
	mark
end

def OFSG.mark(pt, face, vertex, edge, signature=nil)
	mark = OFS_Mark.new
	mark.pt = (pt) ? pt.clone : nil
	mark.face = face
	mark.vertex = vertex
	mark.edge = edge
	mark.signature = signature
	mark
end

#calculate the normal to an edge of a face pointing toward the outside
def OFSG.normal_ex_to_edge(edge, face)
	pt1 = edge.start.position
	pt2 = edge.end.position
	vec = face.normal * pt1.vector_to(pt2)
	vec.length = 1.0
	edge.reversed_in?(face) ? vec : vec.reverse
end

#check whether a point is strictly within a face (edges and vertices excluded)
def OFSG.within_face?(face, pt)
	return (face.classify_point(pt) == 1) if SU_MAJOR_VERSION_6
	pts = []
	face.outer_loop.vertices.each { |v| pts.push v.position }
	Geom.point_in_polygon_2D pt, pts, false
end

#check whether a point is  within a face (edges and vertices included)
def OFSG.within_face_extended?(face, pt)
	if SU_MAJOR_VERSION_6
		case face.classify_point(pt)
		#when 1, 2, 4
		when 1, 2, 4
			return true
		else
			return false
		end
	end	
	pts = []
	face.outer_loop.vertices.each { |v| pts.push v.position }
	return false unless Geom.point_in_polygon_2D pt, pts, true
	return true if face.loops.length == 1
	
	face.loops.each do |loop|
		next if loop.outer?
		pts = []
		loop.vertices.each { |v| pts.push v.position }
		return false if Geom.point_in_polygon_2D pt, pts, true
	end
	return true
end

#Check if 2 points are on the same side of a plane
def OFSG.same_side_of_plane?(plane, pt1, pt2)
	return true if pt1.on_plane?(plane) || pt2.on_plane?(plane)
	pj1 = pt1.project_to_plane plane
	pj2 = pt2.project_to_plane plane
	pj1.vector_to(pt1) % pj2.vector_to(pt2) >= 0
end

end #class OFSG

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# OFSTool: Tool to select a plane, direction and angle - More or less mimic Skecthup Protractor
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
					
class OFSTool

def initialize
	#Loading strings and cursors
	Traductor.load_translation OffsetOnSurface, /MSG_/, binding, "@msg_"
	@idcursor_main = create_cursor OFS_CURSOR_OK
	@idcursor_forbidden = create_cursor OFS_CURSOR_FORBIDDEN
	@idcursor_drag = create_cursor OFS_CURSOR_DRAG
	
	#initializing variables
	@ip_origin = Sketchup::InputPoint.new
	@ip_end = Sketchup::InputPoint.new
	
	#Initializing parameters
	@option_group = false
	@option_alone = false
	@option_nofaces = false
	@option_nocurves = false
	@option_contours = 'O'
	
end

def create_cursor(cursorname)
	cursorpath = Sketchup.find_support_file cursorname, "Plugins/" + OFS_DIR
	cursorpath = Sketchup.find_support_file cursorname, "Plugins" unless cursorpath
	(cursorpath) ? UI::create_cursor(cursorpath, 10, 10) : 0
end

def activate
	@model = Sketchup.active_model
	@selection = @model.selection
	@entities = @model.active_entities
	@ofs = OFS.new
	
	@ofs.set_option_group @option_group
	@ofs.set_option_alone @option_alone
	@ofs.set_option_nofaces @option_nofaces
	@ofs.set_option_nocurves @option_nocurves
	@ofs.set_option_contours @option_contours
	
	@pt_origin = nil
	@pt_end = nil
	@state = STATE_ORIGIN
	@enter_down = false
	@pt_target = nil

	#checking the initial selection
	case @ofs.check_initial_selection(true) 
	when ERROR_INVALID_SELECTION, ERROR_COMPLEX_SELECTION
		@selection.clear
		@noselection = true
	when ERROR_NO_SELECTION	
		@noselection = true
	else	
		@noselection = false
		clean_selection
	end	
		
	info_show
	Sketchup.active_model.active_view.invalidate
		
end

def deactivate(view)
	view.invalidate
end

def getMenu(menu)
	if (@state >= STATE_END)
		menu.add_item(@msg_MnuDone) { call_execute }
		menu.add_separator
	end	
	option_contextual_menu menu
end

def onSetCursor
	if (@selection.length > 0)
		id = @idcursor_main
		id = @idcursor_drag if (@pt_end && @idcursor_drag != 0)
		UI::set_cursor id if (id != 0)	
	else
		UI::set_cursor @idcursor_forbidden if (@idcursor_forbidden != 0)
	end
end

def call_execute
	#d = @pt_target.distance @pt_end
	d = compute_distance
	@ofs.execute d
	@selection.clear if @noselection
	activate
end

def start_over
	@selection.clear
	activate
end

def onLButtonDoubleClick(flags, x, y, view)
	if @state == STATE_EXECUTION
		#call_execute
	else
		#UI.beep
	end
end

def onKeyDown(key, rpt, flags, view)
	key = Traductor.check_key key, flags, false

	case key
		#Toggling between fixed and variable length
		when COPY_MODIFIER_KEY
			return
			
		#Calling options
		when OFS___Group
			change_option_group
		when OFS___Alone
			change_option_alone
		when OFS___NoFaces
			change_option_nofaces
		when OFS___NoCurves
			change_option_nocurves
		when OFS___Contours
			change_option_contours
			
		when 13
			@enter_down = true
	end	
	
	onMouseMove(flags, @xend, @yend, view) if (@state >= STATE_END)
	view.invalidate
	info_show
end

def onKeyUp(key, rpt, flags, view)
	key = Traductor.check_key key, flags, true
	
	case key
	when 13			#Return key
		return unless @enter_down
		@enter_down = false
		return UI.beep if @state < STATE_END
		call_execute
	#Calling dialog box
	when 9		#TAB key
		view.invalidate
		info_show
		return true
	else
		return false
	end	
	true
end

#Control the states of the tool
def set_state(state)
	@state = state
	#@state = STATE_EXECUTION if (@state > STATE_EXECUTION)
	if state == STATE_EXECUTION 
		return call_execute
	end	
	@pt_end = nil if @state == STATE_END
	OFSG.watchlist_clear if @state == STATE_ORIGIN
	
	info_show
end

def onLButtonDown(flags, x, y, view)
	return UI.beep unless @pt_target
	@time_mouse_down = Time.now
	set_state @state + 1
	@ofs.set_red_point(@pt_target, @facetarget, @etarget) if @state == STATE_END
end

def onLButtonUp(flags, x, y, view)
	return unless @pt_target
	if (@state == STATE_ORIGIN)
		return unless @ip_origin.valid?
	elsif (@state == STATE_END)	
		return unless @ip_end.valid? && @pt_end && (@pt_origin != @pt_end)
		delta = Time.now - @time_mouse_down
		dist = (@xend - @xorig) * (@xend - @xorig) + (@yend - @yorig) * (@yend - @yorig)
	    return if (delta < 0.5) && (dist < 100)  
	end
	set_state @state + 1
end

#Handle Escape key and Change of tool
def onCancel(flag, view)
	return  if (flag != 0) || (@state == STATE_ORIGIN)  #Exiting the tool
	set_state @state - 1
end

def clean_selection
	@selection.clear
	@ofs.hsh_faces.each { |key, f| @selection.add f }
	@selection.add @ofs.lst_edges
end

def target_point(view, pt)
	@etarget = nil
	@facetarget = nil
	dtarget = 0.cm
	vptarget = nil
	first = true
	vpt = view.screen_coords pt
	#vpt.z = 0
	
	@ofs.lst_edges.each do |e|
		vpt1 = view.screen_coords e.start.position
		vpt2 = view.screen_coords e.end.position
		#vpt1.z = vpt2.z = 0
		vptproj = vpt.project_to_line [vpt1, vpt2]
		inside = (vptproj == vpt1) || (vptproj == vpt2) || (vpt1.vector_to(vptproj) % vpt2.vector_to(vptproj) < 0)
		vpicked = (inside) ? vptproj : ((vpt.distance(vpt1) < vpt.distance(vpt2)) ? vpt1 : vpt2)
		d = vpt.distance vpicked
		if (first) || (d < dtarget)
			dtarget = d  
			@etarget = e
			@etarget.faces.each do |f|
				if @ofs.hsh_faces[f.to_s]
					@facetarget = f
					break
				end
			end	
			vptarget = vpicked.clone
			first = false
		end
	end
	lp = Geom.closest_points view.pickray(vptarget.x, vptarget.y), @etarget.line 
	lp[1]
end

def onMouseMove(flags, x, y, view)
	case @state	
	when STATE_ORIGIN		#input Origin of Vector
		@ip_origin.pick view, x, y
		break unless @ip_origin.valid?
		@xorig = x
		@yorig = y
		view.tooltip = @ip_origin.tooltip
		@pt_origin = @ip_origin.position if @ip_origin.valid?
		if @noselection
			face = @ip_origin.face
			if face && @entities.include?(face) && @ofs.virtual_selection(face)
				clean_selection unless @selection.include? face
				@pt_target = target_point view, @ip_origin.position
			else
				@selection.clear
				@ofs.reset
				@pt_target = nil
			end
		else	
			@pt_target = target_point view, @ip_origin.position
		end	
		@pt_end = nil

	when STATE_END			#input End of Vector
		@ip_end.pick view, x, y
		break unless @ip_end.valid?
		@xend = x
		@yend = y
		view.tooltip = @ip_end.tooltip
		if @ip_end != @ip_origin
			@vector = edge_perpendicular(@etarget)
			@pt_end = compute_lock(view, flags, @ip_end, @vector)
		end	
	end	
	view.invalidate
	info_show
end

def edge_perpendicular(e)
	ed = @ofs.hsh_edges[e.to_s]
	#vec = ed.face_in.normal * e.start.position.vector_to(e.end.position)
	vec = ed.normal_in
	vec = OFSG.normal_ex_to_edge(@etarget, @facetarget)
end

def compute_distance
	d = @pt_target.distance @pt_end
	vec = @pt_target.vector_to @pt_end
	ed = @ofs.hsh_edges[@etarget.to_s]
	normal = OFSG.normal_ex_to_edge(@etarget, ed.face_in)
	(vec % normal < 0) ? -d : d
end

#Projection of input point for axis lock
def compute_lock(view, flags, ip, vec)
	vdir = @pt_target.vector_to ip.position
	if (!vdir.valid? || vec.parallel?(vdir))
		return ip.position
	elsif (flags == 0) && (ip.degrees_of_freedom == 0)	#When Shift pressed, skip inference
		return ip.position.project_to_line([@pt_target, vec])
	else
		pvorig = view.screen_coords @pt_target
		pv0 = view.screen_coords @pt_target.offset(vec, 100)
		pvip = view.screen_coords ip.position
		pv1 = pvip.project_to_line [pvorig, pv0]
		a = Geom.closest_points [@pt_target, vec], view.pickray(pv1.x, pv1.y)
		return a[0]
	end	
end

#Input of length in the VCB
def onUserText(text, view) 
	@enter_down = false
	begin
		len = text.to_l
		if len == 0
			UI.beep
			return				
		end	
		if @state == STATE_ORIGIN
			if (@pt_target)
				@ofs.execute len
				start_over
			end	
		else
			d = compute_distance
			if len > 0
				d = (d < 0) ? -len : len
			else
				d = (d > 0) ? len : -len
			end	
			@ofs.execute d
			start_over
		end	
	rescue
		UI.beep
	end	
end

#Draw method for tool
def draw(view)
	if (@state >= STATE_ORIGIN)
		if (@pt_target)
			lst_contour = []
			@ofs.lst_edges.each do |e| 
				return start_over unless e.valid?
				p1 = view.screen_coords(e.start.position)
				p2 = view.screen_coords(e.end.position)
				lst_contour.push p1, p2 
			end	
			view.drawing_color = "blue"
			view.line_width = 1
			view.line_stipple = "-"
			view.draw2d GL_LINES, lst_contour
			view.draw_points @pt_target, 6, 2, 'red'
		end
	end
	
	if (@state >= STATE_END && @pt_end)
		d = compute_distance
		view.line_width = 2
		view.drawing_color = "red"
		view.line_stipple = "-"
		if d == 0
			view.draw_lines @pt_target, @pt_end
		else
			path = @ofs.path_from_red_point d
			view.draw GL_LINE_STRIP, path if path && path.length > 1
		end
		if (@ip_end.position != @pt_end)
			view.line_width = 1
			view.drawing_color = "darkgreen"
			view.line_stipple = "-"
			view.draw_lines @ip_end.position, @pt_end
		end
		
		d = compute_distance
		if @option_group
			view.line_stipple = (@option_alone) ? "_" : "-"
			view.line_width = 3
			view.drawing_color = (@option_alone) ? "darkgreen": "darkred"
		else
			view.line_stipple = (@option_alone) ? "_" : ""
			view.line_width = 2
			view.drawing_color = "purple"
		end
		lpt =  @ofs.get_new_contour(d)
		lpt.each { |l| view.draw GL_LINE_STRIP, l if l.length > 0 }
		OFSG.watchlist_draw view
	end
end

def option_contextual_menu(menu)	
	txcur = @msg_MnuCurrent
	text = @msg_MnuContours + " #{txcur} " + Traductor[DLG_EnumContours[@option_contours]] + ") --> F5"
	menu.add_item(text) { self.change_option_contours }
	text = @msg_MnuGroup + " #{txcur} " + Traductor[DLG_EnumYesNo[(@option_group) ? 'Y' : 'N']] + ") --> F6"
	menu.add_item(text) { self.change_option_group }
	text = @msg_MnuAlone + " #{txcur} " + Traductor[DLG_EnumYesNo[(@option_alone) ? 'Y' : 'N']] + ") --> F7"
	menu.add_item(text) { self.change_option_alone }
	text = @msg_MnuNoFaces + " #{txcur} " + Traductor[DLG_EnumYesNo[(@option_nofaces) ? 'Y' : 'N']] + ") --> F8"
	menu.add_item(text) { self.change_option_nofaces }
	text = @msg_MnuNoCurves + " #{txcur} " + Traductor[DLG_EnumYesNo[(@option_nocurves) ? 'Y' : 'N']] + ") --> F9"
	menu.add_item(text) { self.change_option_nocurves }

end

def change_option_group
	@option_group = !@option_group
	@ofs.set_option_group @option_group
end

def change_option_alone
	@option_alone = !@option_alone
	@ofs.set_option_alone @option_alone
end
	
def change_option_nofaces
	@option_nofaces = !@option_nofaces
	@ofs.set_option_nofaces @option_nofaces
end

def change_option_nocurves
	@option_nocurves = !@option_nocurves
	@ofs.set_option_nocurves @option_nocurves
end

def change_option_contours
	case @option_contours 
	when 'O'
		@option_contours = 'A'
	when 'A'
		@option_contours = 'I'
	when 'I'
		@option_contours = 'O'
	end	
	@ofs.set_option_contours @option_contours
	clean_selection
end
	
#display information in the Sketchup status bar
def info_show
	case @state
	when STATE_ORIGIN
		msg = @msg_Face_Origin
	when STATE_END
		msg = @msg_Face_End
	when STATE_EXECUTION
		msg = @msg_Input_Execution
	end
	if (@pt_target && @pt_end)
		d = @pt_target.distance @pt_end
	else	
		d = ""
	end	
	case @option_contours
	when 'A'
		msg += " [" + @msg_Contours + ": " + @msg_AllContours + "]"
	when 'I'
		msg += " [" + @msg_Contours + ": " + @msg_InnerContours + "]"
	when 'O'
		msg += " [" + @msg_Contours + ": " + @msg_OuterContours + "]"
	end	
	msg += " [" + @msg_Group + "]" if (@option_group)
	msg += " [" + @msg_Alone + "]" if (@option_alone)
	msg += " [" + @msg_NoFaces + "]" if (@option_nofaces)
	msg += " [" + @msg_NoCurves + "]" if (@option_nocurves)
		
	Sketchup.set_status_text msg	
	Sketchup.set_status_text @msg_Distance, SB_VCB_LABEL
	Sketchup.set_status_text d, SB_VCB_VALUE
end
	
end #Class OFSTool

#--------------------------------------------------------------------------------------------------------------
# Class ProgressionBar: progress bar in the Sketchup Status text area
#--------------------------------------------------------------------------------------------------------------
class ProgressionBar

#Initialization of progress bar
def initialize(nbelts, label)
	@pb_nbelts = nbelts
	@pb_label = label
	@pb_progression = 0
	@pb_rangemax = 200
	@pb_range = 0
	@pb_time0 = Time.now	
end

#Increment the Progression Bar by <nb> steps
def countage(nb=1)
	@pb_progression += nb
	f = 100 * @pb_progression / @pb_nbelts
	percent = f.to_i
	if (percent != @pb_range)
		@pb_range = percent
		n = 1 + percent * @pb_rangemax / 100
		Sketchup::set_status_text "|" * n.to_i
	end	
	Sketchup.set_status_text @pb_label + " #{@pb_progression} / #{@pb_nbelts}", SB_VCB_LABEL
	Sketchup::set_status_text "#{@pb_range}%  -  #{sprintf "%4.2f", Time.now - @pb_time0} sec", SB_VCB_VALUE
end

end #class ProgressionBar

#--------------------------------------------------------------------------------------------------------------
# Public methods of module Bender
#--------------------------------------------------------------------------------------------------------------

def OffsetOnSurface.set_command
	iconpath = OFS_ICON_MAIN
	menutool = UI.menu OFS_MENU_SU
	menutool.add_separator
	@tlb = nil
	
	#Main command
	cmd = UI::Command.new(Traductor[STR_Title]) { OffsetOnSurface.launch }
	menutool.add_item cmd
	tooltip = Traductor[STR_Tooltip]
	
	cmd.status_bar_text = tooltip
	icon = Sketchup.find_support_file iconpath, "Plugins/" + OFS_DIR
	icon = Sketchup.find_support_file iconpath, "Plugins" unless icon
	if (icon)
		cmd.tooltip = tooltip
		cmd.small_icon = cmd.large_icon = icon
		@tlb = UI::Toolbar.new OFS_TOOLBAR unless @tlb
		@tlb.add_separator
		@tlb.add_item cmd
	end	
	@tlb.show if @tlb
end

unless $OffsetOnSurface____loaded
	OffsetOnSurface.set_command
	$OffsetOnSurface____loaded = true
end
	
end #module OffsetOnSurface
